# Contributing to napari

We welcome your contributions! Please see the provided steps below and never hesitate to contact us.

If you are a new user, we recommend checking out the detailed [Github Guides](https://guides.github.com).

## Setting up a development installation

In order to make changes to `napari`, you will need to [fork](https://guides.github.com/activities/forking/#fork) the
[repository](https://github.com/napari/napari).

If you are not familiar with `git`, we recommend reading up on [this guide](https://guides.github.com/introduction/git-handbook/#basic-git).

Clone the forked repository to your local machine and change directories:
```sh
$ git clone https://github.com/your-username/napari.git
$ cd napari
```

Set the `upstream` remote to the base `napari` repository:
```sh
$ git remote add upstream https://github.com/napari/napari.git
```

Install the required dependencies:
```sh
$ pip install -r requirements.txt
```

Make the development version available globally:
```sh
$ pip install -e .
```

We use
[`pre-commit`](https://pre-commit.com) to run [`black`](https://github.com/psf/black) formatting and [`flake8`](https://github.com/PyCQA/flake8) linting automatically prior to each commit.  Please install it in your environment as follows:
```sh
pre-commit install
```
Upon committing, your code will be formatted according to our [`black` configuration](../pyproject.toml),
which includes the settings `skip-string-normalization = true` and `max-line-length = 79`.
To learn more, see [`black`'s documentation](https://black.readthedocs.io/en/stable/).

Code will also be linted to enforce the stylistic and logistical rules specified in our [`flake8` configuration](../setup.cfg), which currently ignores [E203](https://lintlyci.github.io/Flake8Rules/rules/E203.html), [E501](https://lintlyci.github.io/Flake8Rules/rules/E501.html), [W503](https://lintlyci.github.io/Flake8Rules/rules/W503.html) and [C901](https://lintlyci.github.io/Flake8Rules/rules/C901.html).  For information on any specific flake8 error code, see the [Flake8 Rules](https://lintlyci.github.io/Flake8Rules/).  You may also wish to refer to the [PEP 8 style guide](https://www.python.org/dev/peps/pep-0008/).

If you wish to tell the linter to ignore a specific line use the `# noqa` comment along with the specific error code (e.g. `import sys  # noqa: E402`) but please do not ignore errors lightly.


## Building the icons

Our icon build process starts with a base set of `svg` formatted icons stored in `napari/resources/icons/svg`. If you want to add a new icon to the app, make the icon in whatever program you like, add it to that folder, and make sure it is listed in both `napari/resources/build_icons.py` and `napari/resources/res.qrc`.

Then run the following command

```sh
python -m napari.resources.build_icons
```

This creates separate folders in `napari/resources/icons` with full colored icon sets for each of our themes.

Make sure that the file `napari/resources/icons/res.qrc` has a list of all the icons with the correct prefix for each of our themes, e.g. `<file>icons/light/select.svg</file>`, `<file>icons/dark/select.svg</file>`, etc.

Finally, compile all the icons for use in the app by calling either

```sh
# for PySide2 users
pyside2-rcc -o napari/resources/qt.py napari/resources/res.qrc
```

or

```sh
# for PyQt5 users
pyrcc5 -o napari/resources/qt.py napari/resources/res.qrc
```

depending on if you are using `PySide2` or `PyQt5` respectively. Those two commands are automatically installed when
pip installing the respective packages.

Finally inside the autogenerated `napari/resources/qt.py` replace either `PySide2` or `PyQt5` with `qtpy` to ensure compatibility with both packages.

If you want to change one of the existing icons, modify the version in `napari/resources/icons/svg` and run the commands above.

## Making changes

Create a new feature branch:
```sh
$ git checkout master -b your-branch-name
```

`git` will automatically detect changes to a repository.
You can view them with:
```sh
$ git status
```

Add and commit your changed files:
```sh
$ git add my-file-or-directory
$ git commit -m "my message"
```

### Help us make sure it's you

Each commit you make must have a [GitHub-registered email](https://github.com/settings/emails)
as the `author`. You can read more [here](https://help.github.com/en/github/setting-up-and-managing-your-github-user-account/setting-your-commit-email-address).

To set it, use `git config --global user.email your-address@example.com`.

## Keeping your branches up-to-date

Switch to the `master` branch:
```sh
$ git checkout master
```

Fetch changes and update `master`:
```sh
$ git pull upstream master --tags
```

This is shorthand for:
```sh
$ git fetch upstream master --tags
$ git merge upstream/master
```

Update your other branches:
```sh
$ git checkout your-branch-name
$ git merge master
```

## Sharing your changes

Update your remote branch:
```sh
$ git push -u origin your-branch-name
```

You can then make a [pull-request](https://guides.github.com/activities/forking/#making-a-pull-request) to `napari`'s `master` branch.

## Building the docs

From the project root:
```sh
$ make docs
```

The docs will be built at `docs/build/html`.

Most web browsers will allow you to preview HTML pages.
Try entering `file:///absolute/path/to/napari/docs/build/html/index.html` in your address bar.

## Code of conduct

`napari` has a [Code of Conduct](CODE_OF_CONDUCT.md) that should be honored by everyone who participates in the `napari` community.

## Questions, comments, and feedback

If you have questions, comments, suggestions for improvement, or any other inquiries
regarding the project, feel free to open an [issue](https://github.com/napari/napari/issues).

Issues and pull-requests are written in [Markdown](https://guides.github.com/features/mastering-markdown/#what). You can find a comprehensive guide [here](https://guides.github.com/features/mastering-markdown/#syntax).
